package spire.math

import scala.{specialized => spec}

import spire.algebra._

object Hyper {
  def zero[@spec(Double) A](implicit ev: Fractional[A]) = new Hyper(ev.zero, ev.zero)
  def one[@spec(Double) A](implicit ev: Fractional[A]) = new Hyper(ev.one, ev.zero)

  def ε[@spec(Double) A](implicit ev: Fractional[A]) = new Hyper(ev.one, ev.fromInt(-1))
  def ω[@spec(Double) A](implicit ev: Fractional[A]) = new Hyper(ev.one, ev.one)

  def e[@spec(Double) A](implicit ev: Fractional[A]) = new Hyper(ev.one, ev.fromInt(-1))
  def w[@spec(Double) A](implicit ev: Fractional[A]) = new Hyper(ev.one, ev.one)

  implicit def apply[@spec(Double) A](a: A)(implicit ev: Fractional[A]): Hyper[A] =
    if (ev.eqv(a, ev.zero)) zero else new Hyper(a, ev.zero)

  implicit def order[A](implicit ev: Fractional[A]) = new Order[Hyper[A]] {
    def compare(x: Hyper[A], y: Hyper[A]) =
      x compare y
    def eqv(x: Hyper[A], y: Hyper[A]) =
      ev.eqv(x.degree, y.degree) && ev.eqv(x.n, y.n)
  }

  implicit def field[A](implicit ev: Fractional[A]) = new Field[Hyper[A]] {
    def zero = Hyper.zero[A]
    def one = Hyper.one[A]
    def negate(x: Hyper[A]) = -x
    def plus(x: Hyper[A], y: Hyper[A]) = x + y
    override def minus(x: Hyper[A], y: Hyper[A]) = x - y
    def times(x: Hyper[A], y: Hyper[A]) = x * y
    def div(x: Hyper[A], y: Hyper[A]) = x / y
    override def reciprocal(x: Hyper[A]) = x.reciprocal
    def quot(x: Hyper[A], y: Hyper[A]) = x /~ y
    def mod(x: Hyper[A], y: Hyper[A]) = x % y

    def isWhole(x: Hyper[A]) = x.isWhole
    def ceil(x: Hyper[A]) = x.ceil
    def floor(x: Hyper[A]) = x.floor
    def round(x: Hyper[A]) = x.round

    def gcd(x: Hyper[A], y: Hyper[A]) =
      new Hyper(ev.gcd(x.n, y.n), ev.min(x.degree, y.degree))
  }
}

/**
 * Implements hyperreal numbers in terms of A.
 * 
 * The A type is required to be an ordered field with roots. This type
 * only tracks the most significant part of a hyperreal number. For
 * instance, when adding an infinitesimal number a to a finite number
 * b, the infinitesimal part will be "lost". This makes this type an
 * approximation of a hyperreal number (like Double).
 * 
 * We define ω (or Hyper.w) to be an infinitely large number (larger
 * than any finite number), and then we define ε (or Hyper.e) as 1/ω.
 * 
 * Of course, there are infinite numbers both smaller than ω (like
 * 0.5ω) and larger (like 33ω). There are even numbers infinitely
 * larger than ω like (ω pow 2).
 *
 * In general you can assume that ω and ε behave as one would expect:
 * (2 / ε) == 2ω, (3ε pow 2)(9ω) == 81ε, and so on. Remember though,
 * that like floating point, these approximations are not associative
 * for addition. For example: (ω - ω) + ε == ε, but ω - (ω + ε) == 0.
 */
class Hyper[@spec(Double) A] private[math] (val n: A, val degree: A)(implicit ev: Fractional[A]) { lhs =>
  override def equals(rhs: Any): Boolean =
    rhs match {
      case h: Hyper[_] => degree == h.degree && n == h.n
      case r => isFinite && n == r
    }

  def isInfinite: Boolean = ev.gt(degree, ev.zero)
  def isFinite: Boolean = ev.eqv(degree, ev.zero)
  def isInfinitesimal: Boolean = ev.lt(degree, ev.zero)

  def isZero: Boolean = ev.eqv(n, ev.zero)

  def st: A = if (isFinite) n else ev.zero

  def compare(rhs: Hyper[A]): Int = {
    val sign = ev.compare(lhs.n, rhs.n)
    if (sign == 0) ev.compare(lhs.degree, rhs.degree) else sign
  }

  def <(rhs: Hyper[A]): Boolean = lhs.compare(rhs) < 0
  def <=(rhs: Hyper[A]): Boolean = lhs.compare(rhs) <= 0
  def >(rhs: Hyper[A]): Boolean = lhs.compare(rhs) > 0
  def >=(rhs: Hyper[A]): Boolean = lhs.compare(rhs) >= 0

  def unary_-(): Hyper[A] = new Hyper(ev.negate(n), degree)
  def reciprocal(): Hyper[A] = new Hyper(ev.reciprocal(n), ev.negate(degree))

  def +(rhs: Hyper[A]): Hyper[A] =
    if (lhs.isZero) rhs
    else if (rhs.isZero) lhs
    else if (ev.lt(lhs.degree, rhs.degree)) rhs
    else if (ev.gt(lhs.degree, rhs.degree)) lhs
    else new Hyper(ev.plus(lhs.n, rhs.n), lhs.degree)

  def -(rhs: Hyper[A]): Hyper[A] = lhs + (-rhs)

  def *(rhs: Hyper[A]): Hyper[A] =
    if (lhs.isZero || rhs.isZero) Hyper.zero
    else new Hyper(ev.times(lhs.n, rhs.n), ev.plus(lhs.degree, rhs.degree))

  def /(rhs: Hyper[A]): Hyper[A] =
    if (rhs.isZero) throw new IllegalArgumentException("/ by 0")
    else new Hyper(ev.div(lhs.n, rhs.n), ev.minus(lhs.degree, rhs.degree))

  def /~(rhs: Hyper[A]): Hyper[A] =
    if (rhs.isZero) throw new IllegalArgumentException("/ by 0")
    else if (ev.lt(lhs.degree, rhs.degree)) Hyper.zero
    else new Hyper(ev.quot(lhs.n, rhs.n), ev.minus(lhs.degree, rhs.degree))

  def %(rhs: Hyper[A]): Hyper[A] =
    if (rhs.isZero) throw new IllegalArgumentException("/ by 0")
    else lhs - (lhs /~ rhs) * rhs

  def pow(rhs: Int): Hyper[A] = lhs ** rhs
  def **(rhs: Int): Hyper[A] = new Hyper(ev.pow(lhs.n, rhs), ev.times(lhs.degree, ev.fromInt(rhs)))

  def pow(rhs: A): Hyper[A] = lhs ** rhs
  def **(rhs: A): Hyper[A] = new Hyper(ev.fpow(lhs.n, rhs), ev.times(lhs.degree, rhs))

  def sqrt: Hyper[A] = new Hyper(ev.sqrt(lhs.n), ev.div(lhs.degree, ev.fromInt(2)))

  def nroot(k: Int): Hyper[A] = new Hyper(ev.nroot(lhs.n, k), ev.div(lhs.degree, ev.fromInt(k)))

  def isWhole = isInfinite || isFinite && ev.isWhole(n)

  def ceil: Hyper[A] =
    if (isInfinite) this
    else if (isFinite) new Hyper(ev.ceil(n), ev.zero)
    else if (ev.lt(n, ev.zero)) Hyper.zero
    else Hyper.one

  def floor: Hyper[A] =
    if (isInfinite) this
    else if (isFinite) new Hyper(ev.floor(n), ev.zero)
    else if (ev.lt(n, ev.zero)) -Hyper.one
    else Hyper.zero

  def round: Hyper[A] =
    if (isInfinite) this
    else if (isFinite) new Hyper(ev.round(n), ev.zero)
    else Hyper.zero

  private def suffix: String =
    if (ev.eqv(degree, ev.fromInt(-1))) "ε"
    else if (ev.eqv(degree, ev.one)) "ω"
    else if (isInfinitesimal) "ε^" + ev.negate(degree)
    else if (isInfinite) "ω^" + degree
    else ""

  override def toString(): String = n.toString + suffix
}
