package spire.mathx

import spire.algebra._
import spire.math._

object Flags {
  @inline final def UnboundBelow = 0
  @inline final def OpenBelow = 1
  @inline final def ClosedBelow = 3

  @inline final def UnboundAbove = 0
  @inline final def OpenAbove = 4
  @inline final def ClosedAbove = 12
}

import Flags._

object Interval {
  def empty[A: Order: Ring]: Interval[A] = new Interval(Ring[A].zero, Ring[A].zero, 5)

  def point[A: Order: Ring](a: A): Interval[A] = new Interval(a, a, 15)

  def apply[A: Order: Ring](x: A, y: A): Interval[A] = closed(x, y)

  def closed[A: Order: Ring](x: A, y: A): Interval[A] = {
    if (Order[A].gt(x, y)) throw new IllegalArgumentException("invalid range: %s, %s" format (x, y))
    new Interval(x, y, 15)
  }

  def open[A: Order: Ring](x: A, y: A): Interval[A] = {
    if (Order[A].gt(x, y)) throw new IllegalArgumentException("invalid range: %s, %s" format (x, y))
    new Interval(x, y, 5)
  }

  def unbound[A: Order: Ring]: Interval[A] = new Interval(Ring[A].zero, Ring[A].zero, 0)

  def above[A: Order: Ring](a: A): Interval[A] = new Interval(a, Ring[A].zero, 1)
  def below[A: Order: Ring](a: A): Interval[A] = new Interval(Ring[A].zero, a, 4)

  def atOrAbove[A: Order: Ring](a: A): Interval[A] = new Interval(a, Ring[A].zero, 3)
  def atOrBelow[A: Order: Ring](a: A): Interval[A] = new Interval(Ring[A].zero, a, 12)

  def closedToOpen[A: Order: Ring](x: A, y: A): Interval[A] = {
    if (Order[A].gteqv(x, y)) throw new IllegalArgumentException("invalid range: %s, %s" format (x, y))
    new Interval(x, y, 7)
  }

  def openToClosed[A: Order: Ring](x: A, y: A): Interval[A] = {
    if (Order[A].gteqv(x, y)) throw new IllegalArgumentException("invalid range: %s, %s" format (x, y))
    new Interval(x, y, 13)
  }
}

class Interval[A] private[mathx] (
  private[mathx] val lower: A,
  private[mathx] val upper: A,
  private[mathx] val status: Int
)(implicit o: Order[A], r: Ring[A]) {

  final def isUnboundBelow: Boolean = (status & 3) == 0
  final def isOpenBelow: Boolean = (status & 3) == 1
  final def isClosedBelow: Boolean = (status & 3) == 3

  final def isUnboundAbove: Boolean = (status & 12) == 0
  final def isOpenAbove: Boolean = (status & 12) == 4
  final def isClosedAbove: Boolean = (status & 12) == 12

  def isEmpty: Boolean = status == 5 && o.eqv(lower, upper)
  def isPoint: Boolean = status == 15 && o.eqv(lower, upper)
  def isAbove(a: A): Boolean = !isUnboundBelow && (o.lt(a, lower) || (isOpenBelow && o.eqv(a, lower)))
  def isBelow(a: A): Boolean = !isUnboundAbove && (o.gt(a, upper) || (isOpenAbove && o.eqv(a, upper)))
  def isAtOrAbove(a: A): Boolean = !isUnboundBelow && (o.lteqv(a, lower))
  def isAtOrBelow(a: A): Boolean = !isUnboundAbove && (o.gteqv(a, upper))
  def isAt(a: A): Boolean = status == 15 && o.eqv(a, upper) && o.eqv(a, lower)
  def contains(a: A): Boolean = !isAbove(a) && !isBelow(a)
  def crosses(a: A): Boolean = (isUnboundBelow || o.lt(lower, a)) && (isUnboundAbove || o.lt(a, upper))

  def isBelowZero: Boolean = isBelow(r.zero)
  def crossesZero: Boolean = crosses(r.zero)
  def isAboveZero: Boolean = isAbove(r.zero)

  def mask(that: Interval[A]): Interval[A] = {
    var s: Int = 0
    var l: A = r.zero
    var u: A = r.zero

    var c = o.compare(this.lower, that.lower)
    if (this.isUnboundBelow || c < 0) {
      l = that.lower
      s = that.status & 3
    } else if (that.isUnboundBelow || c > 0) {
      l = this.lower
      s = this.status & 3
    } else {
      l = this.lower
      s = (this.status | that.status) & 3
    }

    c = o.compare(this.upper, that.upper)
    if (this.isUnboundAbove || c > 0) {
      l = that.upper
      s |= that.status & 12
    } else if (that.isUnboundAbove || c < 0) {
      l = this.lower
      s |= this.status & 12
    } else {
      l = this.lower
      s |= (this.status | that.status) & 12
    }

    new Interval(l, u, s)
  }    

  def split(a: A): (Interval[A], Interval[A]) = if (isEmpty || isAt(a))
    (Interval.empty[A], Interval.empty[A])
  else if (isAbove(a))
    (Interval.empty[A], this)
  else if (isBelow(a))
    (this, Interval.empty[A])
  else if (isAtOrAbove(a))
    (Interval.empty[A], new Interval(a, upper, (status & 12) | 1))
  else if (isAtOrBelow(a))
    (new Interval(lower, a, (status & 3) | 4), Interval.empty[A])
  else
    (new Interval(lower, a, (status & 3) | 4), new Interval(a, upper, (status & 12) | 1))

  def splitAtZero: (Interval[A], Interval[A]) = split(r.zero)

  override def toString(): String = if (isEmpty) {
    "∅"
  } else if (isPoint) {
    lower.toString
  } else {
    val s1 = if (isUnboundBelow) "(-∞" else if (isOpenBelow) ("(" + lower) else ("[" + lower)
    val s2 = if (isUnboundAbove) "+∞)" else if (isOpenAbove) (upper.toString + ")") else (upper.toString + "]")
    s1 + ", " + s2
  }

  def unary_-(): Interval[A] = {
    val s = status
    new Interval(r.negate(upper), r.negate(lower), ((s & 3) << 2) | ((s & 12) >> 2))
  }

  def +(that: Interval[A]): Interval[A] = {
    val s = this.status & that.status
    val l = r.plus(this.lower, that.lower)
    val u = r.plus(this.upper, that.upper)
    new Interval(l, u, s)
  }

  def -(that: Interval[A]): Interval[A] = {
    val s2 = that.status
    val s3 = this.status & (((s2 & 3) << 2) | ((s2 & 12) >> 2))
    val l = r.minus(this.lower, that.upper)
    val u = r.minus(this.upper, that.lower)
    new Interval(l, u, s3)
  }

  // def *(that: Interval[A]): Interval[A] = this.status match {
  //   case 0 => // (-inf, inf)
  //     if (that.isEmpty) that else this
  //   case 1 => // (a, inf)

  //   case 3 => // [a, inf)
  //   case 4 => // (-inf, b)
  //   case 5 => // (a, b)
  //   case 7 => // [a, b)
  //   case 12 => // (-inf, b]
  //   case 13 => // (a, b]
  //   case 15 => // [a, b]
  // }

  // def *(that: Interval[A]): Interval[T] = {
  //   val lcz = this.crossesZero
  //   val rcz = that.crossesZero
  
  //   val ll = lower * rhs.lower
  //   val lu = lower * rhs.upper
  //   val ul = upper * rhs.lower
  //   val uu = upper * rhs.upper
  
  //   if (lcz && rcz)
  //     coerce(lu min ul, ll max uu)
  //   else if (lcz)
  //     coerce(ll min lu, ul max uu)
  //   else if (rcz)
  //     coerce(ll min ul, lu max uu)
  //   else if (isBelow(ev.zero) == rhs.isBelow(ev.zero))
  //     coerce(ll min uu, ll max uu)
  //   else
  //     coerce(lu min ul, lu max ul)
  // }

}
